#include "gmQntNorm.h"

#define SIGN_MASK  	0xe000 // upper 3 bits
#define XSIGN_MASK 	0x8000
#define YSIGN_MASK 	0x4000
#define ZSIGN_MASK 	0x2000 
#define X_MASK  		0x1f80 // middle 6 bits
#define Y_MASK  		0x007f // lower 7 bits

gmQuantisedNormal::Scale_lut gmQuantisedNormal::scale_lut;

// Class to allow lut to be initialised automagically.
gmQuantisedNormal::Scale_lut::Scale_lut() 
{
  unsigned int xbits, ybits;
  int idx;
  float x,y,z;

  for (idx = 0; idx < SCALED_LUT_LENGTH; idx++) {
     xbits = (idx & X_MASK) >> 7;
     ybits = (idx & Y_MASK);
        
     if ((xbits + ybits) >= 127 ) {
         xbits = 127 - xbits;
         ybits = 127 - ybits;
     }          // convert to 3D vectors
     x = (float)xbits;
     y = (float)ybits;
     z = (float)(126-xbits-ybits);

     // calculate the amount of normalization required
	  lut[idx] = 1.0f / sqrt(y*y + z*z + x*x);
  }
}

void 
gmQuantisedNormal::quantise(const gmVector3& n)
{
	float scale;
	float norm[3], sum;
	unsigned int xbits,ybits;

	_n = 0;
	n.copyTo(norm);

   if ( norm[0] < 0 ) { _n |= XSIGN_MASK; norm[0] = -norm[0]; }
   if ( norm[1] < 0 ) { _n |= YSIGN_MASK; norm[1] = -norm[1]; }
   if ( norm[2] < 0 ) { _n |= ZSIGN_MASK; norm[2] = -norm[2]; }

	sum = norm[0] + norm[1] + norm[2];
	if (sum == 0) {
		_n = 0;
		return;
	}

	scale = 126.0f/sum;
	xbits = (unsigned int)(norm[0]*scale);
	ybits = (unsigned int)(norm[1]*scale);
	assert(xbits < 127 && xbits >= 0); 
	assert(ybits < 127 && ybits >= 0);  
   if ( xbits >= 64 ) {
         xbits = 127 - xbits;
         ybits = 127 - ybits;
   }       
	// y in the range (0,127) and x in the range (0,63),
   _n |= ((xbits<<7)&X_MASK) | (ybits&Y_MASK);          
}

gmVector3
gmQuantisedNormal::unpack() const
{
   // Extract the points on the plane passing through the unit coordinates
	// Adjust the resulting point to give a unit length.
	// The amount we need to scale by is precomputed! Just a lookup.
   int xbits,ybits;
	gmVector3 result;

   xbits = ((_n & X_MASK) >> 7 );
   ybits = (_n & Y_MASK );       
	
	// Calculate the points on plane defined by ((0,0)(0,126)(126,0))
   if (( xbits + ybits ) >= 127 ) {
			// ubias the points
         xbits = 127 - xbits;
         ybits = 127 - ybits;
   }     
	
   result[0] = (float) (_n & XSIGN_MASK)? -xbits: xbits;
	result[1] = (float) (_n & YSIGN_MASK)? -ybits: ybits;
	result[2] = (float) (126 - xbits - ybits);
		
   if (_n & ZSIGN_MASK) {
		result[2] = -result[2];
	}
		
	return (result*gmQuantisedNormal::scale_lut[_n & ~SIGN_MASK]);
}

#ifdef TEST

#define TEST_RANGE 64
#define TEST_ANGERROR 1.0       
void main() 
{
	 gmVector3 vec2,vec;                
	 gmQuantisedNormal qn,qn2;
		 

	 float maxError = 0;
	 float avgError = 0;
	 int numVecs = 0;       
	 int x,y,z,w;
	 printf("Please wait a few seconds...\n");
	 for ( int x = -TEST_RANGE; x < TEST_RANGE; x++ ) {
			for ( int y = -TEST_RANGE; y < TEST_RANGE; y++ ) {
				  for ( int z = -TEST_RANGE; z < TEST_RANGE; z++ ) {
					 if (( x + y + z ) == 0 ) continue;                
					 vec = gmVector3(x,y,z).normalize();
					 qn = vec;   // quantise
					 vec2 = qn;  // unpack

					 float ang = dot(vec,vec2); // get the angle between them
					 ang = (( fabs( ang ) > 0.9999f ) ? 0 : (float)acos(ang));                

					 if ( ang > TEST_ANGERROR ) {
							cerr << "error: " << ang <<
									  "orig vec: " << vec << ", quant: " << qn << endl; 
					 }
					 avgError += ang;
					 numVecs++;
					 if ( maxError < ang ) maxError = ang;
			     }
			}
	 }       
	 // 0.00 error indicates a insignifcant error
	  printf("max angle error: %f deg, average error: %f deg, num tested vecs: %d\n",
			gmDegrees(maxError), gmDegrees(avgError/numVecs), numVecs);
}
#endif
