/* Extended libgm++ routine - Graphics Math Library
 *
 * Caleb Lyness - 2002
 *
 * Class for describing quantised normals, Usefull for storage and transmission
 * of normals, quantises 3*float normal into 1 16 bit short. 3 bits used for
 * each axis sign, and 6 and 7 bit fields are used for X,Y components, Z is implicit.
 *
 * Lacroute's - Shear Warp Thesis, pg. 159 states that 6 bits for each X and Y 
 * and further 1 bit for the sign of Z is required to prevent "objectional 
 * quantization errors". As we are using 7,8 bits for X and Y (including sign)
 * we are more than covered. 
 */

#ifndef GMQNTNORM_H_CAL
#define GMQNTNORM_H_CAL

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "gmUtils.h"
#include "gmVec3.h"

#define SCALED_LUT_LENGTH 0x2000

class gmVector3;

/* If you want to work with normals, convert to and from a gmVector3 */
class gmQuantisedNormal {
	protected:
		unsigned short _n;

      struct Scale_lut {
			    Scale_lut();
		     	 inline float operator[](int i) const {return lut[i];}
				 float lut[SCALED_LUT_LENGTH];
		};

		static Scale_lut scale_lut;
      
	public:
		gmQuantisedNormal(unsigned short v=0) {_n=v;}
		gmQuantisedNormal(const gmVector3& n) {quantise(n);}
      gmQuantisedNormal(double x, double y, double z) {quantise(gmVector3(x,y,z));}

	   // assignment
      gmQuantisedNormal& assign(double, double, double);
      gmQuantisedNormal& operator =(const gmVector3& rhs); 
      //gmQuantisedNormal& operator =(const gmQuantisedNormal& rhs) {n_ = rhs.n_; return *this; }

		// Type cast to gmVector3
		inline operator gmVector3() const {return unpack();}

		inline unsigned short getQuantisedData(){return _n;}

		inline bool operator == (const gmQuantisedNormal& rhs) const {return (_n == rhs._n);}
		inline bool operator != (const gmQuantisedNormal& rhs) const {return (_n != rhs._n);}

		// output
		friend ostream & operator << ( ostream &, const gmQuantisedNormal & );

	protected:
		void quantise(const gmVector3&);
		gmVector3 unpack() const;
		static void initScaleLUT();
};

// OPERATIONS

inline gmQuantisedNormal& gmQuantisedNormal::assign(double x, double y, double z)
{
   quantise(gmVector3(x,y,z));
	return *this;
}

inline gmQuantisedNormal& gmQuantisedNormal::operator =(const gmVector3& rhs) 
{
		  quantise(rhs); 
		  return *this;
}

// OUTPUT

inline ostream& operator << ( ostream& os, const gmQuantisedNormal& n)
{
  os << (n.unpack());
  return os;
}


#endif
