/******************************************************************************
 * volume_shadow.cpp 
 * implements the construction and drawing of shadow volumes from triangulated surface meshes 
 * Copyright (C) 2010 Marc Gissler <gisslerm@informatik.uni-freiburg.de>
 ******************************************************************************/

#include "volume_shadow.h"

#include "scene.h"
#include "lightmanager.h"

#include <set>

//#define BRUTE_FORCE_SHADOW_VOLUMES

CVolumeShadow::CVolumeShadow()
{
	//m_color = ShadowColor(0.5, 0.5, 0.5, 0.2);
	//clear();
}

CVolumeShadow::~CVolumeShadow()
{
	clear();
}

void CVolumeShadow::build(int lightId, gmVector3& lightPos)
{
	std::map<int, CShadowData*>::iterator mapIter;
	mapIter = m_mapLightAndData.find(lightId);

	if(mapIter == m_mapLightAndData.end())
	{
		CVolumeShadowData* data = new CVolumeShadowData();
		mapIter = m_mapLightAndData.insert(std::pair<int, CShadowData*>(lightId, data)).first;
	}

	CVolumeShadowData* data = static_cast<CVolumeShadowData*>((*mapIter).second);
	std::vector<ShadowFace>& vectorShadowFaces = data->m_vectorShadowFaces;
	std::vector<ShadowVertex>& vectorShadowVertices = data->m_vectorShadowVertices;

	// clear the previous shadow volumes
	vectorShadowFaces.clear();

	std::list<CMesh*>& listMeshes = CScene::getScene()->getMeshes();
	std::list<CMesh*>::iterator meshIter;
	for(meshIter = listMeshes.begin(); meshIter != listMeshes.end(); ++meshIter)
	{
		CMesh* mesh = *meshIter;

		gmMatrix4& transf = mesh->m_transform;

		const std::vector<CMesh::Vertex>& vecVertices = mesh->getVertices();
		std::vector<CMesh::Face>& vecFaces = mesh->getFaces();
		const std::vector<CMesh::Edge>& vecEdges = mesh->getEdges();
		
		// build the projected shadow vertices at infinity
		vectorShadowVertices.reserve(vecVertices.size());
		vectorShadowVertices.resize(vecVertices.size());
		for(size_t vid = 0; vid < vecVertices.size(); ++vid)
		{
			CMesh::Vertex v = transf.transform(vecVertices[vid]);
			
			// projected vertices
			ShadowVertex& vinf = vectorShadowVertices[vid];
			vinf = v; vinf -= lightPos; vinf.w = 0.0;
		}
		
#ifdef BRUTE_FORCE_SHADOW_VOLUMES
		// "brute-force" solution that constructs a shadow volume for each lit face on its own
		for(size_t faceId = 0; faceId < vecFaces.size(); ++faceId)
		{
			const CMesh::Face& face = vecFaces[faceId]; 
			CMesh::Vertex v1 = transf.transform(vecVertices[face.v1]);
			CMesh::Vertex v2 = transf.transform(vecVertices[face.v2]);
			CMesh::Vertex v3 = transf.transform(vecVertices[face.v3]);

			// compute incident light direction ( IncidentLightDir = AverageFacePosition - LightPosition)
			//ATTENTION: these positions MUST NOT have w != 1.0!
			gmVector3 avgPos(v1 + v2 + v3); avgPos /= 3.0;
			gmVector3 lightDir = avgPos - lightPos;

			// if the primitive is facing away from the light source, continue
			if(dot(lightDir, face.normal) >= 0.0) continue;

			// we compute transformation in double precision, which leads to other results than the single-precision floating-point transformations on the GPU
			// therefore, we add a little offset in the opposite direction of the face normal to avoid z-fighting in the stencil-tests
			v1 += face.normal * -0.001;
			v2 += face.normal * -0.001;
			v3 += face.normal * -0.001;

			// projected vertices
			ShadowVertex& v0inf = vectorShadowVertices[face.v1];
			ShadowVertex& v1inf = vectorShadowVertices[face.v2];
			ShadowVertex& v2inf = vectorShadowVertices[face.v3];

			// add the face of the mesh that faces the light source, too
			{
				ShadowFace shadowFace = ShadowFace(ShadowVertex(v1), ShadowVertex(v2), ShadowVertex(v3));
				vectorShadowFaces.push_back(shadowFace);
			}

			// add the "bottom" of the shadow volume
			{
				ShadowFace shadowFace = ShadowFace(v0inf, v2inf, v1inf);
				vectorShadowFaces.push_back(shadowFace);
			}

			// --------------------------------------------------
			// construct two triangles for each of face edges
			//

			//edge0
			{
				ShadowFace shadowFace0 = ShadowFace(ShadowVertex(v1), v1inf, ShadowVertex(v2));
				vectorShadowFaces.push_back(shadowFace0);

				ShadowFace shadowFace1 = ShadowFace(ShadowVertex(v1), v0inf, v1inf);
				vectorShadowFaces.push_back(shadowFace1);
			}

			//edge1
			{
				ShadowFace shadowFace0 = ShadowFace(ShadowVertex(v2), v2inf, ShadowVertex(v3));
				vectorShadowFaces.push_back(shadowFace0);

				ShadowFace shadowFace1 = ShadowFace(ShadowVertex(v2), v1inf, v2inf);
				vectorShadowFaces.push_back(shadowFace1);
			}

			//edge2
			{
				ShadowFace shadowFace0 = ShadowFace(ShadowVertex(v3), v0inf, ShadowVertex(v1));
				vectorShadowFaces.push_back(shadowFace0);

				ShadowFace shadowFace1 = ShadowFace(ShadowVertex(v3), v2inf, v0inf);
				vectorShadowFaces.push_back(shadowFace1);
			}
		}
#else
		// solution that produces a lower fill rate by only constructing a shadow volume based on extruded silhouette edges
		
		// compute wheter a face is lit by this light source or not
		for(size_t faceId = 0; faceId < vecFaces.size(); ++faceId)
		{
			CMesh::Face& face = vecFaces[faceId]; 
			CMesh::Vertex v1 = transf.transform(vecVertices[face.v1]);
			CMesh::Vertex v2 = transf.transform(vecVertices[face.v2]);
			CMesh::Vertex v3 = transf.transform(vecVertices[face.v3]);

			// compute incident light direction ( IncidentLightDir = AverageFacePosition - LightPosition)
			//ATTENTION: these positions MUST NOT have w != 1.0!
			gmVector3 avgPos(v1 + v2 + v3); avgPos /= 3.0;
			gmVector3 lightDir = avgPos - lightPos;

			// if the primitive is facing away from the light source, continue
			if(dot(lightDir, face.normal) < 0.0)
			{
				face.bLit = true;
				
				// we compute transformation in double precision, which leads to other results than the single-precision floating-point transformations on the GPU
				// therefore, we add a little offset in the opposite direction of the face normal to avoid z-fighting in the stencil-tests
				v1 += face.normal * -0.001;
				v2 += face.normal * -0.001;
				v3 += face.normal * -0.001;
				
				// projected vertices
				ShadowVertex& v0inf = vectorShadowVertices[face.v1];
				ShadowVertex& v1inf = vectorShadowVertices[face.v2];
				ShadowVertex& v2inf = vectorShadowVertices[face.v3];

				// add the face of the mesh that faces the light source, too
				{
					ShadowFace shadowFace = ShadowFace(ShadowVertex(v1), ShadowVertex(v2), ShadowVertex(v3));
					vectorShadowFaces.push_back(shadowFace);
				}

				// add the "bottom" of the shadow volume
				{
					ShadowFace shadowFace = ShadowFace(v0inf, v2inf, v1inf);
					vectorShadowFaces.push_back(shadowFace);
				}
			}
			else
			{
				face.bLit = false;
			}
		}

		for(size_t edgeId = 0; edgeId < vecEdges.size(); ++edgeId)
		{
			const CMesh::Edge& edge = vecEdges[edgeId];

			if(edge.neightborFace[0] == -1 ||edge.neightborFace[1] == -1) continue;

			// skip, if the edge is no silhouette edge (both are lit or both are in shadow)
			if(vecFaces[edge.neightborFace[0]].bLit == vecFaces[edge.neightborFace[1]].bLit) continue;

			CMesh::Face& face = (vecFaces[edge.neightborFace[0]].bLit) ? vecFaces[edge.neightborFace[0]] : vecFaces[edge.neightborFace[1]];
			
			int sum = 0;
			if(face.v1 == edge.v0Id || face.v1 == edge.v1Id) sum +=1;
			if(face.v2 == edge.v0Id || face.v2 == edge.v1Id) sum +=2;
			if(face.v3 == edge.v0Id || face.v3 == edge.v1Id) sum +=3;

			int index0, index1;
			if(sum == 3)
			{
				index0 = face.v1; 
				index1 = face.v2;
			}
			else if(sum == 5)
			{
				index0 = face.v2;
				index1 = face.v3;
			}
			else if(sum == 4)
			{
				index0 = face.v3;
				index1 = face.v1;
			}
			else
				printf("error!\n");

			CMesh::Vertex v1 = transf.transform(vecVertices[index0]);
			CMesh::Vertex v2 = transf.transform(vecVertices[index1]);
			v1 += face.normal * -0.001;
			v2 += face.normal * -0.001;
			
			ShadowVertex& v0inf = vectorShadowVertices[index0];
			ShadowVertex& v1inf = vectorShadowVertices[index1];
			
			{
				ShadowFace shadowFace0 = ShadowFace(ShadowVertex(v1), v1inf, ShadowVertex(v2));
				vectorShadowFaces.push_back(shadowFace0);

				ShadowFace shadowFace1 = ShadowFace(ShadowVertex(v1), v0inf, v1inf);
				vectorShadowFaces.push_back(shadowFace1);
			}
		}
#endif

	}
}

void CVolumeShadow::clear()
{
	std::map<int, CShadowData*>::iterator mapIter;
	for(mapIter = m_mapLightAndData.begin(); mapIter != m_mapLightAndData.end(); ++mapIter)
	{
		CVolumeShadowData* data = static_cast<CVolumeShadowData*>(mapIter->second);
		delete data;
	}
	m_mapLightAndData.clear();
}

void CVolumeShadow::m_drawShadowVolumes(int lightId)
{
	std::map<int, CShadowData*>::iterator mapIter;
	mapIter = m_mapLightAndData.find(lightId);

	if(mapIter == m_mapLightAndData.end())
		return; // error: should be in here!
	
	CVolumeShadowData* data = static_cast<CVolumeShadowData*>((*mapIter).second);
	std::vector<ShadowFace>& vectorShadowFaces = data->m_vectorShadowFaces;

	// draw the shadow primitives
	glBegin(GL_TRIANGLES);
	for(size_t i = 0; i < vectorShadowFaces.size(); ++i)
	{
		ShadowFace& face = vectorShadowFaces[i];

		//note: give the vertices as 4D-vector to be able to determine the W-coordinate (some of the vertices are located at inifinity (w = 0) )
		glVertex4f( (GLfloat) (face.v1.x), (GLfloat) (face.v1.y), (GLfloat) (face.v1.z), (GLfloat) face.v1.w);
		glVertex4f( (GLfloat) (face.v2.x), (GLfloat) (face.v2.y), (GLfloat) (face.v2.z), (GLfloat) face.v2.w);
		glVertex4f( (GLfloat) (face.v3.x), (GLfloat) (face.v3.y), (GLfloat) (face.v3.z), (GLfloat) face.v3.w);
	}
	glEnd();
}

void CVolumeShadow::drawSceneWithShadows()
{
	CScene* pScene = CScene::getScene();
	std::map<int, CShadowData*>::iterator mapIter;

	for(mapIter = m_mapLightAndData.begin(); mapIter != m_mapLightAndData.end(); ++mapIter)
	{
		int lightId = mapIter->first;
		glEnable(GL_LIGHT0+lightId);
	}

	// 1st pass: render ambient intensities
	CLightManager::getLightManager()->setAmbientIntensity();
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	pScene->drawMeshes();
	
	// render all intensities but the ambient
	CLightManager::getLightManager()->setAllIntensitiesButAmbient();

	for(mapIter = m_mapLightAndData.begin(); mapIter != m_mapLightAndData.end(); ++mapIter)
	{
		int lightId = mapIter->first;
		glDisable(GL_LIGHT0+lightId);
	}

	// perform the z-fail algorithm for each light separately
	for(mapIter = m_mapLightAndData.begin(); mapIter != m_mapLightAndData.end(); ++mapIter)
	{
		int lightId = mapIter->first;
		glEnable(GL_LIGHT0+lightId);

		glEnable(GL_STENCIL_TEST);
		glEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);

		glDepthMask(GL_FALSE);
		glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

		glStencilFunc(GL_ALWAYS, 0x00000000, 0x00000000);

		// 2nd pass	
		
		glStencilOp(GL_KEEP, GL_INCR, GL_KEEP);
		glCullFace(GL_FRONT);
		m_drawShadowVolumes(lightId);

		// 3rd pass
		
		glStencilOp(GL_KEEP, GL_DECR, GL_KEEP );
		glCullFace(GL_BACK);

		m_drawShadowVolumes(lightId);

		// everything back to normal
		glDepthMask(GL_TRUE);
		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
		glCullFace(GL_BACK);
		glDepthFunc(GL_LEQUAL);
		glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

		//// set shadow color
		//CLight* l0 = CLightManager::getLightManager()->getLights()[0];
		//glColor4f(l0->m_shadowColor[0], l0->m_shadowColor[1], l0->m_shadowColor[2], 0.5f);

		// draw shadowed part (not necessary, since we already rendered the ambient light intensities to the color buffer in the 1st pass

		// draw lit part	
		
		glStencilFunc(GL_EQUAL, 0x00000000, 0xffffffff);
		glEnable(GL_BLEND);
		glBlendFunc(GL_ONE, GL_ONE);
		pScene->drawMeshes();	
		glDisable(GL_BLEND);
		glDepthFunc(GL_LESS);
		glDisable(GL_CULL_FACE);
		glDisable(GL_STENCIL_TEST);
		glClear(GL_STENCIL_BUFFER_BIT);
		glDisable(GL_LIGHT0+lightId);
	}

}
